#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""The constraints defining if a swift display_item is assumed a NxImageRealSpace concept."""

# pylint: disable=no-member,line-too-long

# releasing line-too-long restriction to avoid having line breaks in the mapping table
# made the experience that when having a widescreen working with the mapping table
# as single-line instructions is more convenient to read and parsable by human eye

# current issues:
# hard-coding currently the ebeam deflector is problematic as it is
# to just make the name variadic but then assume there is always only one
# in the specific case of a Nion microscope and which information is available typically
# via swift there one would get information about the scanbox for which the closest
# match is NXebeam_deflector and there is only one scanbox at an Nion Hermes 200
# but one could customize this implementation here and instead resolve how the scanbox
# is composed of individual groups of deflection coils in which case one could also
# think about using one NXebeam_deflector for every single coil...

NxImageRealSpaceDict = {"IGNORE": {"fun": "load_from", "terms": "type"},
                        "IGNORE": {"fun": "load_from", "terms": "uuid"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/start_time": {"fun": "convert_iso8601", "terms": ["created", "timezone"]},
                        "IGNORE": {"fun": "load_from", "terms": "is_sequence"},
                        "IGNORE": {"fun": "load_from", "terms": "intensity_calibration/offset"},
                        "IGNORE": {"fun": "load_from", "terms": "intensity_calibration/scale"},
                        "IGNORE": {"fun": "load_from", "terms": "intensity_calibration/units"},
                        "IGNORE": {"fun": "load_from", "terms": "dimensional_calibrations"},
                        "IGNORE": {"fun": "load_from", "terms": "timezone"},
                        "IGNORE": {"fun": "load_from", "terms": "timezone_offset"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/high_tension"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/defocus"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/electron_source/voltage": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/EHT"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/electron_source/voltage/@units": "V",
                        "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/PMTBF_gain"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/PMTDF_gain"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/STAGE_LAB[stage_lab]/tilt1": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/StageOutA"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/STAGE_LAB[stage_lab]/tilt1/@units": "deg",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/STAGE_LAB[stage_lab]/tilt2": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/StageOutB"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/STAGE_LAB[stage_lab]/tilt2/@units": "deg",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/STAGE_LAB[stage_lab]/position": {"fun": "load_from", "terms": ["metadata/instrument/ImageScanned/StageOutX", "metadata/instrument/ImageScanned/StageOutY", "metadata/instrument/ImageScanned/StageOutZ"]},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/STAGE_LAB[stage_lab]/position/@units": "m",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_1_0/magnitude": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C10"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_1_0/magnitude/@units": "m",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_1_2_a/magnitude": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C12.a"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_1_2_a/magnitude/@units": "m",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_1_2_b/magnitude": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C12.b"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_1_2_b/magnitude/@units": "m",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_2_1_a/magnitude": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C21.a"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_2_1_a/magnitude/@units": "m",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_2_1_b/magnitude": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C21.b"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_2_1_b/magnitude/@units": "m",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_2_3_a/magnitude": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C23.a"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_2_3_a/magnitude/@units": "m",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_2_3_b/magnitude": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C23.b"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_2_3_b/magnitude/@units": "m",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_3_0/magnitude": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C30"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_3_0/magnitude/@units": "m",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_3_2_a/magnitude": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C32.a"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_3_2_a/magnitude/@units": "m",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_3_2_b/magnitude": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C32.b"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_3_2_b/magnitude/@units": "m",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_3_4_a/magnitude": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C34.a"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_3_4_a/magnitude/@units": "m",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_3_4_b/magnitude": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C34.b"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_3_4_b/magnitude/@units": "m",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_5_0/magnitude": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C50"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/aberration_correction/ZEMLIN_TABLEAU/PROCESS[process]/nion/c_5_0/magnitude/@units": "m",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/LENS_EM[lens_em1]/value": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C1 ConstW"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/LENS_EM[lens_em1]/name": "C1",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/LENS_EM[lens_em2]/value": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C2 ConstW"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/LENS_EM[lens_em2]/name": "C2",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/LENS_EM[lens_em3]/value": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/C3 ConstW"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_COLUMN[ebeam_column]/LENS_EM[lens_em3]/name": "C3",
                        "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/PMT2_gain"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/SuperFEG.^EmissionCurrent"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/G_2Db"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/LastTuneCurrent"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/OPTICAL_SYSTEM_EM[optical_system_em]/semi_convergence_angle": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/probe_ha"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/OPTICAL_SYSTEM_EM[optical_system_em]/semi_convergence_angle/@units": "mrad",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/IMAGE_SET[image_set*]/inner_half_angle": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/HAADF_Inner_ha"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/IMAGE_SET[image_set*]/inner_half_angle/@units": "mrad",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/IMAGE_SET[image_set*]/outer_half_angle": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/HAADF_Outer_ha"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/IMAGE_SET[image_set*]/outer_half_angle/@units": "mrad",
                        "IGNORE": {"fun": "load_from", "terms": "metadata/instrument/ImageScanned/GeometricProbeSize"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/hardware_source_id"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/hardware_source_name"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_id"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/center": {"fun": "load_from", "terms": ["metadata/scan/center_x_nm", "metadata/scan/center_y_nm"]},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/center/@units": "nm",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/OPTICAL_SYSTEM_EM[optical_system_em]/field_of_view": {"fun": "load_from", "terms": "metadata/scan/fov_nm"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/OPTICAL_SYSTEM_EM[optical_system_em]/field_of_view/@units": "nm",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/rotation": {"fun": "load_from", "terms": "metadata/scan/rotation"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/rotation/@units": "deg",
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/rotation_deg"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_context_size"},
                        "UNCLEAR": {"fun": "load_from", "terms": "metadata/scan/subscan_fractional_size"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_size"},
                        "UNCLEAR": {"fun": "load_from", "terms": "metadata/scan/subscan_fractional_center"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/size"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/center_nm"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/pixel_time_us"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/fov_nm"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/rotation_rad"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/external_clock_wait_time": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/external_clock_wait_time_ms"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/external_clock_wait_time": "ms",
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/external_clock_mode": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/external_clock_mode"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/external_scan_mode": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/external_scan_mode"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/external_scan_ratio": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/external_scan_ratio"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/ac_line_sync"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/ac_frame_sync": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/ac_frame_sync"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/flyback_time_us"},
                        "UNCLEAR": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/subscan_pixel_size"},
                        "UNCLEAR": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/subscan_fractional_size"},
                        "UNCLEAR": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/subscan_fractional_center"},
                        "UNCLEAR": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/top_left_override"},
                        "UNCLEAR": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/data_shape_override"},
                        "UNCLEAR": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/state_override"},
                        "UNCLEAR": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/section_rect"},
                        "UNCLEAR": {"fun": "load_from", "terms": "metadata/scan/scan_device_parameters/scan_id"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/ac_line_sync": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/ac_line_sync"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/calibration_style": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/calibration_style"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/center_x_nm"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/center_y_nm"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/flyback_time": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/flyback_time_us"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/flyback_time/@units": "µs",
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/fov_nm"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/line_time": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/line_time_us"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/line_time/@units": "µs",
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/pixel_time_us"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/pixels_x"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/pixels_y"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/pixel_time_target": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/requested_pixel_time_us"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/pixel_time_target/@units": "µs",
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/rotation_deg"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/rotation_rad"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board1]/DAC[dac1]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 0 DAC 0"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board1]/DAC[dac2]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 0 DAC 1"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board1]/DAC[dac3]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 0 DAC 2"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board1]/DAC[dac4]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 0 DAC 3"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board1]/DAC[dac5]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 0 DAC 4"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board1]/DAC[dac6]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 0 DAC 5"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board1]/DAC[dac7]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 0 DAC 6"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board1]/DAC[dac8]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 0 DAC 7"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board1]/DAC[dac9]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 0 DAC 8"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board1]/DAC[dac10]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 0 DAC 9"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board1]/DAC[dac11]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 0 DAC 10"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board1]/DAC[dac12]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 0 DAC 11"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board1]/relay": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 0 Relay"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board2]/DAC[dac1]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 1 DAC 0"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board2]/DAC[dac2]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 1 DAC 1"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board2]/DAC[dac3]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 1 DAC 2"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board2]/DAC[dac4]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 1 DAC 3"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board2]/DAC[dac5]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 1 DAC 4"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board2]/DAC[dac6]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 1 DAC 5"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board2]/DAC[dac7]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 1 DAC 6"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board2]/DAC[dac8]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 1 DAC 7"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board2]/DAC[dac9]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 1 DAC 8"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board2]/DAC[dac10]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 1 DAC 9"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board2]/DAC[dac11]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 1 DAC 10"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board2]/DAC[dac12]/value": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 1 DAC 11"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/em_lab/EBEAM_DEFLECTOR[ebeam_deflector1]/CIRCUIT_BOARD[mag_board2]/relay": {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 1 Relay"},
                        "UNCLEAR": {"fun": "load_from", "terms": "metadata/scan/valid_rows"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/IMAGE_SET[image_set*]/detector_identifier": {"fun": "load_from", "terms": "metadata/hardware_source/hardware_source_id"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/hardware_source_name"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/exposure"},
                        "UNCLEAR": {"fun": "load_from", "terms": "metadata/hardware_source/frame_index"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/channel_id"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/channel_name"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/pixel_time_us"},
                        "IGNORE": {"fun": "load_from", "terms": "metadata/hardware_source/line_time_us"},
                        "UNCLEAR": {"fun": "load_from", "terms": "metadata/hardware_source/valid_rows"},
                        "UNCLEAR": {"fun": "load_from", "terms": "metadata/hardware_source/channel_index"},
                        "UNCLEAR": {"fun": "load_from", "terms": "metadata/hardware_source/reference_key"},
                        "UNCLEAR": {"fun": "load_from", "terms": "metadata/hardware_source/view_id"},
                        "IGNORE": {"fun": "load_from", "terms": "title"},
                        "IGNORE": {"fun": "load_from", "terms": "session_id"},
                        "IGNORE": {"fun": "load_from", "terms": "session"},
                        "IGNORE": {"fun": "load_from", "terms": "category"},
                        "IGNORE": {"fun": "load_from", "terms": "version"},
                        "IGNORE": {"fun": "load_from", "terms": "modified"},
                        "IGNORE": {"fun": "load_from", "terms": "data_shape"},
                        "IGNORE": {"fun": "load_from", "terms": "data_dtype"},
                        "IGNORE": {"fun": "load_from", "terms": "collection_dimension_count"},
                        "IGNORE": {"fun": "load_from", "terms": "datum_dimension_count"},
                        "/ENTRY[entry*]/measurement/EVENT_DATA_EM[event_data_em*]/end_time": {"fun": "convert_iso8601", "terms": ["data_modified", "timezone"]}}



#!/usr/bin/env python
"""
Writes the simplest NeXus HDF5 file using h5py

Uses method accepted at 2014NIAC
according to the example from Figure 1.3
in the Introduction chapter
"""

from pathlib import Path
import h5py
import numpy

filename = str(Path(__file__).absolute().parent.parent / "simple_example.dat")
buffer = numpy.loadtxt(filename).T
tthData = buffer[0]  # float[]
countsData = numpy.asarray(buffer[1], "int32")  # int[]

with h5py.File("simple_example_write1.hdf5", "w") as f:  # create the HDF5 NeXus file
    # since this is a simple example, no attributes are used at this point

    nxentry = f.create_group("Scan")
    nxentry.attrs["NX_class"] = "NXentry"

    nxdata = nxentry.create_group("data")
    nxdata.attrs["NX_class"] = "NXdata"
    nxdata.attrs["signal"] = "counts"
    nxdata.attrs["axes"] = "two_theta"
    nxdata.attrs["two_theta_indices"] = [
        0,
    ]

    tth = nxdata.create_dataset("two_theta", data=tthData)
    tth.attrs["units"] = "degrees"

    counts = nxdata.create_dataset("counts", data=countsData)
    counts.attrs["units"] = "counts"



#!/usr/bin/env python3
"""Main file of yaml2nxdl tool.
Users create NeXus instances by writing a YAML file
which details a hierarchy of data/metadata elements

"""
# -*- coding: utf-8 -*-
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


# Yaml library does not except the keys (escapechar "\t" and yaml separator ":")
# So the corresponding value is to skip them and
# and also carefull about this order
import hashlib
from yaml.composer import Composer
from yaml.constructor import Constructor

from yaml.nodes import ScalarNode
from yaml.resolver import BaseResolver
from yaml.loader import Loader

# NOTE: If any one change one of the bellow dict please change it for both
ESCAPE_CHAR_DICT_IN_YAML = {"\t": "    ",
                            "\':\'": ":"}

ESCAPE_CHAR_DICT_IN_XML = {"    ": "\t",
                           "\':\'": ":"}


class LineLoader(Loader):  # pylint: disable=too-many-ancestors
    """
    LineLoader parses a yaml into a python dictionary extended with extra items.
    The new items have as keys __line__<yaml_keyword> and as values the yaml file line number
    """

    def compose_node(self, parent, index):
        # the line number where the previous token has ended (plus empty lines)
        node = Composer.compose_node(self, parent, index)
        node.__line__ = self.line + 1
        return node

    def construct_mapping(self, node, deep=False):
        node_pair_lst = node.value
        node_pair_lst_for_appending = []

        for key_node in node_pair_lst:
            shadow_key_node = ScalarNode(
                tag=BaseResolver.DEFAULT_SCALAR_TAG, value='__line__' + key_node[0].value)
            shadow_value_node = ScalarNode(
                tag=BaseResolver.DEFAULT_SCALAR_TAG, value=key_node[0].__line__)
            node_pair_lst_for_appending.append(
                (shadow_key_node, shadow_value_node))

        node.value = node_pair_lst + node_pair_lst_for_appending
        return Constructor.construct_mapping(self, node, deep=deep)


def get_yaml_escape_char_dict():
    """Get escape char and the way to skip them in yaml."""
    return ESCAPE_CHAR_DICT_IN_YAML


def get_yaml_escape_char_reverter_dict():
    """To revert yaml escape char in xml constructor from yaml."""

    return ESCAPE_CHAR_DICT_IN_XML


def type_check(nx_type):
    """
        Check for nexus type if type is NX_CHAR get '' or get as it is.
    """

    if nx_type in ['NX_CHAR', '']:
        nx_type = ''
    else:
        nx_type = f"({nx_type})"
    return nx_type


def get_node_parent_info(tree, node):
    """
    Return tuple of (parent, index) where:
    parent = node of parent within tree
    index = index of node under parent
    """

    parent_map = {c: p for p in tree.iter() for c in p}
    parent = parent_map[node]
    return parent, list(parent).index(node)


def cleaning_empty_lines(line_list):
    """
        Cleaning up empty lines on top and bottom.
    """
    if not isinstance(line_list, list):
        line_list = line_list.split('\n') if '\n' in line_list else ['']

    # Clining up top empty lines
    while True:
        if line_list[0].strip():
            break
        line_list = line_list[1:]
        if len(line_list) == 0:
            line_list.append('')
            return line_list

    # Clining bottom empty lines
    while True:
        if line_list[-1].strip():
            break
        line_list = line_list[0:-1]
        if len(line_list) == 0:
            line_list.append('')
            return line_list

    return line_list


def nx_name_type_resolving(tmp):
    """
    extracts the eventually custom name {optional_string}
    and type {nexus_type} from a YML section string.
    YML section string syntax: optional_string(nexus_type)
    """
    if tmp.count('(') == 1 and tmp.count(')') == 1:
        # we can safely assume that every valid YML key resolves
        # either an nx_ (type, base, candidate) class contains only 1 '(' and ')'
        index_start = tmp.index('(')
        index_end = tmp.index(')', index_start + 1)
        typ = tmp[index_start + 1:index_end]
        nam = tmp.replace('(' + typ + ')', '')
        return nam, typ

    # or a name for a member
    typ = ''
    nam = tmp
    return nam, typ


def get_sha256_hash(file_name):
    """Generate a sha256_hash for a given file.
    """
    sha_hash = hashlib.sha256()

    with open(file=file_name, mode='rb',) as file_obj:
        # Update hash for each 4k block of bytes
        for b_line in iter(lambda: file_obj.read(4096), b""):
            sha_hash.update(b_line)
    return sha_hash.hexdigest()


def extend_yamlfile_with_comment(yaml_file,
                                 file_to_be_appended,
                                 top_lines_list=None):
    """Extend yaml file by the file_to_be_appended as comment.
    """

    with open(yaml_file, mode='a+', encoding='utf-8') as f1_obj:
        if top_lines_list:
            for line in top_lines_list:
                f1_obj.write(line)

        with open(file_to_be_appended, mode='r', encoding='utf-8') as f2_obj:
            lines = f2_obj.readlines()
            for line in lines:
                f1_obj.write(f"# {line}")


def separate_hash_yaml_and_nxdl(yaml_file, sep_yaml, sep_xml):
    """Separate the provided yaml file into yaml, nxdl and hash if yaml was extended with
    nxdl at the end of yaml by
        '\n# ++++++++++++++++++++++++++++++++++ SHA HASH \
            ++++++++++++++++++++++++++++++++++\n'
         # <has value>'
    """
    sha_hash = ''
    with open(yaml_file, 'r', encoding='utf-8') as inp_file:
        lines = inp_file.readlines()
        # file to write yaml part
        with open(sep_yaml, 'w', encoding='utf-8') as yml_f_ob, \
                open(sep_xml, 'w', encoding='utf-8') as xml_f_ob:

            last_line = ''
            write_on_yaml = True
            for ind, line in enumerate(lines):
                if ind == 0:
                    last_line = line
                # Write in file when ensured that the nest line is not with '++ SHA HASH ++'
                elif '++ SHA HASH ++' not in line and write_on_yaml:
                    yml_f_ob.write(last_line)
                    last_line = line
                elif '++ SHA HASH ++' in line:
                    write_on_yaml = False
                    last_line = ''
                elif not write_on_yaml and not last_line:
                    # The first line of xml file has been found. Onward write lines directly
                    # into xml file.
                    if not sha_hash:
                        sha_hash = line.split('# ', 1)[-1].strip()
                    else:
                        xml_f_ob.write(line[2:])
            # If the yaml fiile does not contain any hash for nxdl then we may have last line.
            if last_line:
                yml_f_ob.write(last_line)

    return sha_hash



from .anchor_list import AnchorRegistry  # noqa F401
from .nxdl import NXClassDocGenerator  # noqa F401
from .nxdl_index import nxdl_indices  # noqa F401
from .xsd import XSDDocGenerator  # noqa F401



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Compute hashes of files for provenance tracking of data sources."""

import hashlib


def get_file_hashvalue(file_name: str) -> str:
    """Compute a hashvalue of given file, here SHA256."""
    sha256_hash = hashlib.sha256()
    try:
        with open(file_name, "rb") as file_handle:
            # Read and update hash string value in blocks of 4K
            for byte_block in iter(lambda: file_handle.read(4096), b""):
                sha256_hash.update(byte_block)
    except IOError:
        print(f"File {file_name} is not accessible !")

    return sha256_hash.hexdigest()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Representation of an NXevent_data_em class."""

# pylint: disable=no-member

from typing import Dict

import hyperspy.api as hs

from pynxtools.dataconverter.readers.em_spctrscpy.utils.em_nexus_base_classes \
    import NxObject

from pynxtools.dataconverter.readers.em_spctrscpy.utils.hspy.em_hspy_xray \
    import NxSpectrumSetEmXray

from pynxtools.dataconverter.readers.em_spctrscpy.utils.hspy.em_hspy_eels \
    import NxSpectrumSetEmEels

from pynxtools.dataconverter.readers.em_spctrscpy.utils.hspy.em_hspy_adf \
    import NxImageSetEmAdf

from pynxtools.dataconverter.file_hashing import get_file_hashvalue


class NxEventDataEm:
    """Representing a data collection event with a single detector.

    During this data collection event, the microscope
    was considered stable enough.
    """

    def __init__(self, file_name: str, entry_id: int):
        self.file_name = ""
        self.file_sha256 = ""
        self.entry_id = entry_id
        self.meta: Dict[str, NxObject] = {}
        self.meta["start_time"] = NxObject("non_recoverable")
        self.meta["end_time"] = NxObject("non_recoverable")
        # the following list is not complete
        # but brings an example how NXem can be used disentangle
        # data and processing when, at a given point in time,
        # multiple detectors have been used

        if (file_name != "") and (entry_id > 0):
            self.file_name = file_name
            self.file_sha256 = get_file_hashvalue(self.file_name)

            self.spectrum_set_em_xray = NxSpectrumSetEmXray([])
            self.spectrum_set_em_eels = NxSpectrumSetEmEels([])
            self.image_set_em_adf = NxImageSetEmAdf([])

            self.parse_hspy_analysis_results()

    def parse_hspy_analysis_results(self):
        """Parse the individual hyperspy-specific data.

        Route these respective classes of an NxEventDataEm instance.
        """
        hspy_objs = hs.load(self.file_name)
        # this logic is too simplistic e.g. what if the dataset is TBs?
        # because file_name is usually the file from the microscope session...

        # developers can here switch easily one and off certain sub-parsers
        if isinstance(hspy_objs, list):
            print("Processing a list of hspy_objs...")
            self.spectrum_set_em_xray = NxSpectrumSetEmXray(hspy_objs)
            self.spectrum_set_em_eels = NxSpectrumSetEmEels(hspy_objs)
            self.image_set_em_adf = NxImageSetEmAdf(hspy_objs)
        else:
            print("Converting (a single hspy obj) into a list, processing it...")
            self.spectrum_set_em_xray = NxSpectrumSetEmXray([hspy_objs])
            self.spectrum_set_em_eels = NxSpectrumSetEmEels([hspy_objs])
            self.image_set_em_adf = NxImageSetEmAdf([hspy_objs])

    def report(self, template: dict) -> dict:
        """Copy data from self into template the appdef instance.

        Paths in template are prefixed by prefix and have to be compliant
        with the application definition.
        """
        prefix = f"/ENTRY[entry{self.entry_id}]/measurement/" \
                 f"EVENT_DATA_EM[event_data_em1]/"

        # now = datetime.datetime.now().astimezone().isoformat()
        # hyperspy cannot implement per-event time stamping especially
        # not for time-zones because time data are vendor-specifically formatted
        # not always reported in which case hspy replaces missing vendor timestamps
        # with system time at runtime of the script !

        template[f"{prefix}start_time"] = self.meta["start_time"].value
        template[f"{prefix}end_time"] = self.meta["end_time"].value
        event_info = {"source_file_name": self.file_name,
                      "source_file_version": self.file_sha256}

        prefix = f"/ENTRY[entry{self.entry_id}]/measurement/" \
                 f"EVENT_DATA_EM[event_data_em1]/"
        # connect and compare frame_id with that of hspy
        if self.spectrum_set_em_xray is not None:
            if isinstance(self.spectrum_set_em_xray,
                          NxSpectrumSetEmXray) is True:
                self.spectrum_set_em_xray.report(
                    prefix, 1, event_info, template)

        prefix = f"/ENTRY[entry{self.entry_id}]/measurement/" \
                 f"EVENT_DATA_EM[event_data_em1]/"
        if self.spectrum_set_em_eels is not None:
            if isinstance(self.spectrum_set_em_eels,
                          NxSpectrumSetEmEels) is True:
                self.spectrum_set_em_eels.report(
                    prefix, 1, event_info, template)

        prefix = f"/ENTRY[entry{self.entry_id}]/measurement/" \
                 f"EVENT_DATA_EM[event_data_em1]/"
        # connect and compare frame_id with that of hspy
        if self.image_set_em_adf is not None:
            if isinstance(self.image_set_em_adf,
                          NxImageSetEmAdf) is True:
                self.image_set_em_adf.report(
                    prefix, 1, event_info, template)

        # add generic images

        return template



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Resize images so that they display properly in H5Web."""

# pylint: disable=no-member

# f"https://stackoverflow.com/questions/4321290/"
# f" how-do-i-make-pil-take-into-account-the-shortest-side-when-creating-a-thumbnail"

import numpy as np


def thumbnail(img, size=300):
    """Create a thumbnail, i.e. resized version of an image."""
    img = img.copy()

    if img.mode not in ('L', 'RGB'):
        img = img.convert('RGB')

    old_width, old_height = img.size

    if old_width < size and old_height < size:
        return img

    if old_width == old_height:
        img.thumbnail((size, size))
    elif old_height > old_width:
        ratio = float(old_width) / float(old_height)
        new_width = ratio * size
        img = img.resize((int(np.floor(new_width)), size))
    elif old_width > old_height:
        ratio = float(old_height) / float(old_width)
        new_height = ratio * size
        img = img.resize((size, int(np.floor(new_height))))

    return img



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Parse display_items inside a zip file generated from compressing a swift project dir."""

# pylint: disable=no-member

# from typing import Dict, Any, List, Tuple

import yaml

import json

import flatdict as fd

import nion.swift.model.NDataHandler as nsnd

import numpy as np

import h5py

from zipfile37 import ZipFile

from pynxtools.dataconverter.readers.em_nion.utils.swift_uuid_to_file_name \
    import uuid_to_file_name

from pynxtools.dataconverter.readers.em_nion.utils.swift_generate_dimscale_axes \
    import get_list_of_dimension_scale_axes

from pynxtools.dataconverter.readers.em_nion.map_concepts.swift_display_items_to_nx \
    import nexus_concept_dict, identify_nexus_concept_key

from pynxtools.dataconverter.readers.shared.map_concepts.mapping_functors \
    import apply_modifier, variadic_path_to_specific_path

from pynxtools.dataconverter.readers.em_nion.map_concepts.swift_to_nx_image_real_space \
    import NxImageRealSpaceDict

from pynxtools.dataconverter.readers.em_nion.utils.em_nion_versioning \
    import NX_EM_NION_SWIFT_NAME, NX_EM_NION_SWIFT_VERSION
from pynxtools.dataconverter.readers.em_nion.utils.em_nion_versioning \
    import NX_EM_NION_EXEC_NAME, NX_EM_NION_EXEC_VERSION


class NxEmNionSwiftProjectParser:
    """Parse NionSwift project file.

    """

    def __init__(self, file_name, entry_id):
        """Class wrapping swift parser."""
        self.file_name = file_name
        self.entry_id = entry_id
        # counters which keep track of how many instances of NXevent_data_em have
        # been instantiated, this implementation currently maps each display_items
        # onto an own NXevent_data_em instance
        self.event_data_written = False
        self.event_data_em_id = 1
        self.image_id = 1
        self.spectrum_id = 1
        self.proj_file_names = []
        # assure that there is exactly one *.nsproj file only to parse from
        self.ndata_file_dict = {}
        # just get the *.ndata files irrespective whether they will be parsed later
        self.hdf_file_dict = {}
        # just get the *.h5 files irrespective whether they will be interpreted

    def check_project_file(self):
        """Inspect the content of the compressed project file to check if supported."""
        # file_name = "2022-02-18_Metadata_Kuehbach.zip.nionswift"
        with ZipFile(self.file_name) as zip_file_hdl:
            for file in zip_file_hdl.namelist():
                if file.endswith(".h5"):
                    key = file[file.rfind("/") + 1:].replace(".h5", "")
                    if key not in self.hdf_file_dict:
                        self.hdf_file_dict[key] = file
                elif file.endswith(".ndata"):
                    key = file[file.rfind("/") + 1:].replace(".ndata", "")
                    if key not in self.ndata_file_dict:
                        self.ndata_file_dict[key] = file
                elif file.endswith(".nsproj"):
                    self.proj_file_names.append(file)
                else:
                    continue
        if not self.ndata_file_dict.keys().isdisjoint(self.hdf_file_dict.keys()):
            print("Keys of *.ndata and *.h5 files in project are not disjoint!")
            return False
        if len(self.proj_file_names) != 1:
            print("The project contains either no or more than one nsproj file!")
            return False
        print(self.proj_file_names)
        for key, val in self.ndata_file_dict.items():
            print(f"{key}, {val}")
        for key, val in self.hdf_file_dict.items():
            print(f"{key}, {val}")
        return True

    def add_nx_image_real_space(self, meta, arr, template):
        """Add data and metadata for an instance of concept NxImageRealSpace."""
        # metadata
        identifier = [self.entry_id, self.event_data_em_id, self.image_id]
        for nx_path, modifier in NxImageRealSpaceDict.items():
            if (nx_path != "IGNORE") and (nx_path != "UNCLEAR"):
                # print(nx_path)
                # instance_identifier = list(np.repeat(1, nx_path.count("*")))
                # print(instance_identifier)
                trg = variadic_path_to_specific_path(nx_path, identifier)
                # print(trg)
                template[trg] = apply_modifier(modifier, meta)

        # array data
        axes_lst = get_list_of_dimension_scale_axes(meta)
        # print(axes_lst)

        axes_names = [("axis_image_identifier", "image_identifier", 2),
                      ("axis_y", "y", 1),
                      ("axis_x", "x", 0)]
        print(f"Add NXdata len(axes_lst) {len(axes_lst)}, len(axes_names) {len(axes_names)}")
        if 2 <= len(axes_lst) <= len(axes_names):
            trg = f"/ENTRY[entry{self.entry_id}]/measurement/EVENT_DATA_EM[event_data_em" \
                  f"{self.event_data_em_id}]/IMAGE_SET[image_set{self.image_id}]/" \
                  f"PROCESS[process]"
            template[f"{trg}/source"] = "n/a"
            template[f"{trg}/source/@version"] = "n/a"
            template[f"{trg}/PROGRAM[program1]/program"] \
                = f"We do not know because the nsproj file does not store it explicitly "\
                  f"which nionswift version and dependencies are used when writing "\
                  f"the nsproj file!"
            template[f"{trg}/PROGRAM[program1]/program/@version"] = "not recoverable"
            template[f"{trg}/PROGRAM[program2]/program"] \
                = f"{NX_EM_NION_SWIFT_NAME}"
            template[f"{trg}/PROGRAM[program2]/program/@version"] \
                = f"{NX_EM_NION_SWIFT_VERSION}"
            template[f"{trg}/PROGRAM[program3]/program"] \
                = f"{NX_EM_NION_EXEC_NAME}"
            template[f"{trg}/PROGRAM[program3]/program/@version"] \
                = f"{NX_EM_NION_EXEC_VERSION}"

            trg = f"/ENTRY[entry{self.entry_id}]/measurement/EVENT_DATA_EM[event_data_em" \
                  f"{self.event_data_em_id}]/IMAGE_SET[image_set{self.image_id}]/DATA[stack]"
            template[f"{trg}/@NX_class"] = "NXdata"  # ##TODO one should not need to add this manually
            template[f"{trg}/title"] = str("Should come from NionSwift directly")
            template[f"{trg}/@signal"] = "data_counts"
            template[f"{trg}/@axes"] = ["axis_image_identifier", "axis_y", "axis_x"]
            for idx in np.arange(0, 3):
                template[f"{trg}/@AXISNAME_indices[{axes_names[idx][0]}_indices]"] \
                    = np.uint32(axes_names[idx][2])
            # the following three lines would be required by H5Web to plot RGB maps
            # template[f"{trg}/@CLASS"] = "IMAGE"
            # template[f"{trg}/@IMAGE_VERSION"] = "1.2"
            # template[f"{trg}/@SUBCLASS_VERSION"] = np.int64(15)

            if len(axes_lst) == 2:
                ny, nx = np.shape(arr)
                template[f"{trg}/data_counts"] \
                    = {"compress": np.reshape(arr, (1, ny, nx), order="C"), "strength": 1}
                template[f"{trg}/data_counts/@long_name"] = "Signal"
                # no image_identifier axis available
                template[f"{trg}/AXISNAME[{axes_names[0][0]}]"] \
                    = {"compress": np.asarray([1], np.uint32), "strength": 1}
                template[f"{trg}/AXISNAME[{axes_names[0][0]}]/@long_name"] \
                    = f"Image identifier (a. u.)"
                template[f"{trg}/AXISNAME[{axes_names[0][0]}]/@units"] = ""
                for idx in [1, 2]:
                    template[f"{trg}/AXISNAME[{axes_names[idx][0]}]"] \
                        = {"compress": axes_lst[idx - 1]["value"], "strength": 1}
                    template[f"{trg}/AXISNAME[{axes_names[idx][0]}]/@long_name"] \
                        = f"Calibrated position along {axes_names[idx][1]}-axis " \
                          f"({axes_lst[idx - 1]['unit']})"
                    template[f"{trg}/AXISNAME[{axes_names[idx][0]}]/@units"] \
                        = f"{axes_lst[idx - 1]['unit']}"
            else:  # len(axes_lst) == 3
                template[f"{trg}/data_counts"] = {"compress": arr, "strength": 1}
                for idx in [0, 1, 2]:
                    # TODO check that casting works properly
                    template[f"{trg}/AXISNAME[{axes_names[idx][0]}]"] \
                        = {"compress": np.asarray(axes_lst[idx]["value"], np.uint32),
                           "strength": 1}
                    template[f"{trg}/AXISNAME[{axes_names[idx][0]}]/@long_name"] \
                        = f"Calibrated position along {axes_names[idx][1]}-axis " \
                          f"({axes_lst[idx]['unit']})"
                    template[f"{trg}/AXISNAME[{axes_names[idx][0]}]/@units"] \
                        = f"{axes_lst[idx]['unit']}"

        self.image_id += 1
        self.event_data_written = True
        return template

    def update_event_identifier(self):
        """Advance and reset bookkeeping of event data em and data instances."""
        if self.event_data_written is True:
            self.event_data_em_id += 1
            self.event_data_written = False
        self.image_id = 1
        self.spectrum_id = 1
        # because either we found that the display item is fed from an H5 or from an NDATA
        # print(f"Identifier at {self.entry_id}, {self.event_data_em_id}, {self.image_id}, {self.spectrum_id}")

    def map_to_nexus(self, meta, arr, concept_name, template):
        """Create the actual instance of a specific set of NeXus concepts in template."""
        # meta is an flatdict
        # arr is a numpy array
        if concept_name == "NxImageSetRealSpace":
            print(f"Adding an instance of concept {concept_name}")
            self.add_nx_image_real_space(meta, arr, template)
        else:
            print(f"Ignoring concept {concept_name} because not yet implemented")

        self.update_event_identifier()
        return template

    def process_ndata(self, file_hdl, full_path, template):
        """Handle reading and processing of opened *.ndata inside the ZIP file."""
        # assure that we start reading that file_hdl/pointer from the beginning...
        file_hdl.seek(0)
        local_files, dir_files, eocd = nsnd.parse_zip(file_hdl)
        # ...now that pointer might point somewhere...
        flat_metadata_dict = {}
        data_arr = None
        nx_concept_name = ""

        for offset, tpl in local_files.items():
            # print(f"{tpl}")
            if tpl[0] == b'metadata.json':
                print(f"Extract metadata.json from {full_path} at offset {offset}")
                # ... explicit jump back to beginning of the file
                file_hdl.seek(0)
                metadata_dict = nsnd.read_json(file_hdl,
                                               local_files,
                                               dir_files,
                                               b'metadata.json')

                nx_concept_key = identify_nexus_concept_key(metadata_dict)
                nx_concept_name = nexus_concept_dict[nx_concept_key]
                print(f"Display_item {full_path}, concept {nx_concept_key}, maps {nx_concept_name}")

                flat_metadata_dict = fd.FlatDict(metadata_dict, delimiter='/')
                break
                # because we expect (based on Benedikt's example) to find only one json file
                # in that *.ndata file pointed to by file_hdl

        if flat_metadata_dict == {}:  # only continue if some metadata were retrieved
            return template

        for offset, tpl in local_files.items():
            # print(f"{tpl}")
            if tpl[0] == b'data.npy':
                print(f"Extract data.npy from {full_path} at offset {offset}")
                file_hdl.seek(0)
                data_arr = nsnd.read_data(file_hdl,
                                          local_files,
                                          dir_files,
                                          b'data.npy')
                break
                # because we expect (based on Benedikt's example) to find only one npy file
                # in that *.ndata file pointed to by file_hdl

        print(f"data_arr type {data_arr.dtype}, shape {np.shape(data_arr)}")
        # check on the integriety of the data_arr array that it is not None or empty
        # this should be done more elegantly by just writing the
        # data directly into the template and not creating another copy

        self.map_to_nexus(flat_metadata_dict, data_arr, nx_concept_name, template)
        del flat_metadata_dict
        del data_arr
        del nx_concept_name
        return template

    def process_hdf(self, file_hdl, full_path, template):
        """Handle reading and processing of opened *.h5 inside the ZIP file."""
        flat_metadata_dict = {}
        data_arr = None
        nx_concept_name = ""

        file_hdl.seek(0)
        h5r = h5py.File(file_hdl, "r")
        metadata_dict = json.loads(h5r["data"].attrs["properties"])

        nx_concept_key = identify_nexus_concept_key(metadata_dict)
        nx_concept_name = nexus_concept_dict[nx_concept_key]
        print(f"Display_item {full_path}, concept {nx_concept_key}, maps {nx_concept_name}")

        flat_metadata_dict = fd.FlatDict(metadata_dict, delimiter='/')

        if flat_metadata_dict == {}:  # only continue if some metadata were retrieved
            return template

        data_arr = h5r["data"][()]
        h5r.close()

        print(f"data_arr type {data_arr.dtype}, shape {np.shape(data_arr)}")
        # check on the integriety of the data_arr array that it is not None or empty
        # this should be done more elegantly by just writing the
        # data directly into the template and not creating another copy
        self.map_to_nexus(flat_metadata_dict, data_arr, nx_concept_name, template)
        del flat_metadata_dict
        del data_arr
        del nx_concept_name
        return template

    def parse_project_file(self, template: dict) -> dict:
        """Parse lazily from compressed NionSwift project (nsproj + directory)."""
        swift_proj_dict = {}
        with ZipFile(self.file_name) as zip_file_hdl:
            with zip_file_hdl.open(self.proj_file_names[0]) as file_hdl:
                # with open(file_name, 'r') as stream:
                swift_proj_dict = fd.FlatDict(yaml.safe_load(file_hdl), delimiter='/')
                # for entry in swift_proj_dict["display_items"]:
                #     if isinstance(entry, dict):
                #         for key, val in entry.items():
                #             print(f"{key}, {val}")
        if swift_proj_dict == {}:
            return template

        for itm in swift_proj_dict["display_items"]:
            if set(["type", "uuid", "created", "display_data_channels"]).issubset(itm.keys()):
                if len(itm["display_data_channels"]) == 1:
                    if "data_item_reference" in itm["display_data_channels"][0].keys():
                        key = uuid_to_file_name(
                            itm["display_data_channels"][0]["data_item_reference"])
                        # file_name without the mime type
                        if key in self.ndata_file_dict:
                            print(f"Key {key} is *.ndata maps to {self.ndata_file_dict[key]}")
                            with ZipFile(self.file_name) as zip_file_hdl:
                                print(f"Parsing {self.ndata_file_dict[key]}...")
                                with zip_file_hdl.open(self.ndata_file_dict[key]) as file_hdl:
                                    self.process_ndata(
                                        file_hdl,
                                        self.ndata_file_dict[key],
                                        template)
                        elif key in self.hdf_file_dict:
                            print(f"Key {key} is *.h5 maps to {self.hdf_file_dict[key]}")
                            with ZipFile(self.file_name) as zip_file_hdl:
                                print(f"Parsing {self.hdf_file_dict[key]}...")
                                with zip_file_hdl.open(self.hdf_file_dict[key]) as file_hdl:
                                    self.process_hdf(
                                        file_hdl,
                                        self.hdf_file_dict[key],
                                        template)
                        else:
                            print(f"Key {key} has no corresponding data file")
        return template

    def parse(self, template: dict) -> dict:
        """Parse NOMAD OASIS relevant data and metadata from swift project."""
        print("Parsing lazily from compressed NionSwift project (nsproj + directory)...")
        print(self.file_name)
        print(f"{self.entry_id}")
        if self.check_project_file() is False:
            return template

        self.parse_project_file(template)
        return template



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""An example reader implementation for the DataConverter."""
from typing import Tuple, Any, Callable, Dict, List
import os

from pynxtools.dataconverter.readers.base.reader import BaseReader
from pynxtools.dataconverter.template import Template


class YamlJsonReader(BaseReader):
    """A reader that takes a mapping json file and a data file/object to return a template."""

    # pylint: disable=too-few-public-methods

    # Whitelist for the NXDLs that the reader supports and can process
    supported_nxdls: List[str] = []
    extensions: Dict[str, Callable[[Any], dict]] = {}
    kwargs: Dict[str, Any] = None

    def read(
        self,
        template: dict = None,
        file_paths: Tuple[str] = None,
        objects: Tuple[Any] = None,
        **kwargs,
    ) -> dict:
        """
        Reads data from multiple files and passes them to the appropriate functions
        in the extensions dict.
        """
        template = Template()
        self.kwargs = kwargs

        sorted_paths = sorted(file_paths, key=lambda f: os.path.splitext(f)[1])
        for file_path in sorted_paths:
            extension = os.path.splitext(file_path)[1].lower()
            if extension not in self.extensions:
                print(
                    f"WARNING: "
                    f"File {file_path} has an unsupported extension, ignoring file."
                )
                continue
            if not os.path.exists(file_path):
                print(f"WARNING: File {file_path} does not exist, ignoring entry.")
                continue

            template.update(
                self.extensions.get(extension, lambda _: {})(file_path)
            )

        template.update(self.extensions.get("default", lambda _: {})(""))
        template.update(self.extensions.get("objects", lambda _: {})(objects))

        return template


READER = YamlJsonReader



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Classes representing groups with NeXus-ish formatted data parsed from hspy."""

# pylint: disable=no-member

from typing import Dict

import numpy as np

import hyperspy.api as hs

from pynxtools.dataconverter.readers.em_spctrscpy.utils.em_nexus_base_classes \
    import NxObject


class HspyRectRoiXrayAllSpectra:
    """Representing a regular stack of X-ray spectra over a rectangular ROI."""

    def __init__(self, hspy_clss):
        self.meta: Dict[str, NxObject] = {}
        self.meta["program"] = NxObject(value="hyperspy")
        self.meta["program_version"] = NxObject(value=hs.__version__)
        self.meta["title"] = NxObject()
        self.meta["long_name"] = NxObject()
        self.meta["counts"] = NxObject()
        self.meta["xpos"] = NxObject()
        self.meta["xpos_long_name"] = NxObject()
        self.meta["ypos"] = NxObject()
        self.meta["ypos_long_name"] = NxObject()
        self.meta["photon_energy"] = NxObject()
        self.meta["photon_energy_long_name"] = NxObject()
        self.is_valid = True

        self.is_supported(hspy_clss)
        self.parse(hspy_clss)

    def is_supported(self, hspy_s3d):
        """Check if the input has supported axes_manager and key metadata."""
        assert hspy_s3d.metadata["Signal"]["signal_type"] == "EDS_TEM", \
            "hspy_s3d is not a valid hyperspy generic instance !"
        assert hspy_s3d.data.ndim == 3, \
            "hspy_s3d is not a valid 3D dataset !"
        axes_dict = hspy_s3d.axes_manager.as_dictionary()
        required_axis_names = ["axis-0", "axis-1", "axis-2"]
        for req_key in required_axis_names:
            assert req_key in axes_dict.keys(), \
                req_key + " is unexpectedly not registered in the axes_manager !"
        required_keywords = ["_type", "name", "units", "size", "scale", "offset"]
        avail_axis_names = []
        for keyword in axes_dict.keys():
            for req_key in required_keywords:  # check if all required keys exist
                assert req_key in axes_dict[keyword].keys(), \
                    "hspy_s3d axis " + keyword + " lacks " + req_key + " !"

            assert axes_dict[keyword]["_type"] == "UniformDataAxis", \
                keyword + ", this axis is not of type UniformDataAxis !"
            avail_axis_names.append(axes_dict[keyword]["name"])

        axes_as_expected_emd = np.all(
            np.sort(avail_axis_names) == np.sort(["y", "x", "X-ray energy"]))
        axes_as_expected_bcf = np.all(
            np.sort(avail_axis_names) == np.sort(["height", "width", "Energy"]))
        if (axes_as_expected_emd is False) and (axes_as_expected_bcf is True):
            print(f"\tIn function {__name__} as expected")
            self.is_valid = False

    def parse(self, hspy_s3d):
        """Parse a hyperspy Signal2D stack instance into an NX default plottable."""
        if self.is_valid is False:
            pass
        print(f"\tIn function {__name__}")
        self.meta["title"].value = hspy_s3d.metadata["General"]["title"]
        # self.long_name.value = hspy_s3d.metadata["Signal"]["signal_type"]
        self.meta["long_name"].value = hspy_s3d.metadata["General"]["title"]
        self.meta["counts"].value = hspy_s3d.data  # hspy uses numpy and adapts ??
        axes_dict = hspy_s3d.axes_manager.as_dictionary()
        for keyword in axes_dict.keys():
            offset = np.float64(axes_dict[keyword]["offset"])
            scale = np.float64(axes_dict[keyword]["scale"])
            size = np.uint32(axes_dict[keyword]["size"])
            unit = str(axes_dict[keyword]["units"])

            y_axis = (axes_dict[keyword]["name"] == "y") \
                or (axes_dict[keyword]["name"] == "height")
            x_axis = (axes_dict[keyword]["name"] == "x") \
                or (axes_dict[keyword]["name"] == "width")
            e_axis = (axes_dict[keyword]["name"] == "X-ray energy") \
                or (axes_dict[keyword]["name"] == "Energy")
            if y_axis is True:
                self.meta["ypos"].value = np.asarray(
                    np.linspace(0., np.float64(size) * scale, num=size,
                                endpoint=True) + offset / 2., np.float64)
                self.meta["ypos"].unit = unit
                self.meta["ypos_long_name"].value = "y"  # ##MK::name y always!
            if x_axis is True:
                self.meta["xpos"].value = np.asarray(
                    np.linspace(0., np.float64(size) * scale, num=size,
                                endpoint=True) + offset / 2., np.float64)
                self.meta["xpos"].unit = unit
                self.meta["xpos_long_name"].value = "x"  # ##MK::name x always!
            if e_axis is True:
                self.meta["photon_energy"].value = np.asarray(
                    np.linspace(0., np.float64(size) * scale, num=size,
                                endpoint=True) + offset / 2., np.float64)
                self.meta["photon_energy"].unit = unit
                self.meta["photon_energy_long_name"].value = "Energy"
                # ##MK::name Energy always!


class HspyRectRoiXraySummarySpectrum:
    """Representing the accumulated X-ray spectrum over a rectangular ROI."""

    def __init__(self, hspy_clss):
        self.meta: Dict[str, NxObject] = {}
        self.meta["program"] = NxObject(value="hyperspy")
        self.meta["program_version"] = NxObject(value=hs.__version__)
        self.meta["title"] = NxObject()
        self.meta["long_name"] = NxObject()
        self.meta["counts"] = NxObject()
        self.meta["photon_energy"] = NxObject()
        self.meta["photon_energy_long_name"] = NxObject()
        self.is_valid = True

        self.is_supported(hspy_clss)
        self.parse(hspy_clss)

    def is_supported(self, hspy_s1d):
        """Check if the input has supported axes_manager and key metadata."""
        assert hspy_s1d.metadata["Signal"]["signal_type"] == "EDS_TEM", \
            "hspy_s3d is not a valid hyperspy generic instance !"
        assert hspy_s1d.data.ndim == 1, \
            "hspy_s3d is not a valid 1D dataset !"
        axes_dict = hspy_s1d.axes_manager.as_dictionary()
        required_axis_names = ["axis-0"]
        for req_key in required_axis_names:
            assert req_key in axes_dict.keys(), \
                req_key + " is unexpectedly not registered in the axes_manager !"
        required_keywords = ["_type", "name", "units", "size", "scale", "offset"]
        avail_axis_names = []
        for keyword in axes_dict.keys():
            for req_key in required_keywords:  # check if all required keys exist
                assert req_key in axes_dict[keyword].keys(), \
                    "hspy_s1d axis " + keyword + " lacks " + req_key + " !"

            assert axes_dict[keyword]["_type"] == "UniformDataAxis", \
                keyword + ", this axis is not of type UniformDataAxis !"
            avail_axis_names.append(axes_dict[keyword]["name"])

        axes_as_expected = np.all(
            np.sort(avail_axis_names) == np.sort(["Energy"]))
        if axes_as_expected is False:
            print(f"\tIn function {__name__} as expected")
            self.is_valid = False

    def parse(self, hspy_s1d):
        """Parse a hyperspy Signal1D stack instance into an NX default plottable."""
        if self.is_valid is False:
            pass
        print(f"\tIn function {__name__}")
        self.meta["title"].value = hspy_s1d.metadata["General"]["title"]
        # self.meta["long_name"].value = hspy_s1d.metadata["Signal"]["signal_type"]
        self.meta["long_name"].value = hspy_s1d.metadata["General"]["title"]
        self.meta["counts"].value = hspy_s1d.data
        # ##MK::it seems that hspy is adaptive, uses numpy under the hood
        # though, so the .data instance is already a proper numpy dtype
        # therefore, an explicit call like this
        # np.asarray(hspy_s1d.data, np.uint32) is not necessary
        axes_dict = hspy_s1d.axes_manager.as_dictionary()
        for keyword in axes_dict.keys():
            offset = np.float64(axes_dict[keyword]["offset"])
            scale = np.float64(axes_dict[keyword]["scale"])
            size = np.uint32(axes_dict[keyword]["size"])
            unit = str(axes_dict[keyword]["units"])
            # if axes_dict[keyword]["name"] == "Energy":
            self.meta["photon_energy"].value = np.asarray(
                np.linspace(0., np.float64(size) * scale, num=size,
                            endpoint=True) + offset / 2., np.float64)
            self.meta["photon_energy"].unit = unit
            self.meta["photon_energy_long_name"].value = "Energy"


class HspyRectRoiXrayMap:
    """Representing an X-ray composition map with metadata."""

    def __init__(self, hspy_clss):
        self.meta: Dict[str, NxObject] = {}
        self.meta["program"] = NxObject(value="hyperspy")
        self.meta["program_version"] = NxObject(value=hs.__version__)
        self.meta["title"] = NxObject()
        self.meta["long_name"] = NxObject()
        self.meta["counts"] = NxObject()
        self.meta["xpos"] = NxObject()
        self.meta["xpos_long_name"] = NxObject()
        self.meta["ypos"] = NxObject()
        self.meta["ypos_long_name"] = NxObject()
        self.is_valid = True

        self.is_supported(hspy_clss)
        self.parse(hspy_clss)

    def is_supported(self, hspy_s2d):
        """Check if the input has supported axes_manager and key metadata."""
        assert hspy_s2d.metadata["Signal"]["signal_type"] == "", \
            "hspy_s2d is not a valid hyperspy generic instance !"
        assert hspy_s2d.data.ndim == 2, \
            "hspy_s2d is not a valid 2D dataset !"
        axes_dict = hspy_s2d.axes_manager.as_dictionary()
        required_axis_names = ["axis-0", "axis-1"]
        for req_key in required_axis_names:
            assert req_key in axes_dict.keys(), \
                req_key + " is unexpectedly not registered in the axes_manager !"
        required_keywords = ["_type", "name", "units", "size", "scale", "offset"]
        avail_axis_names = []
        for keyword in axes_dict.keys():
            for req_key in required_keywords:  # check if all required keys exist
                assert req_key in axes_dict[keyword].keys(), \
                    "hspy_s2d axis " + keyword + " lacks " + req_key + " !"
            assert axes_dict[keyword]["_type"] == "UniformDataAxis", \
                keyword + ", this axis is not of type UniformDataAxis !"
            avail_axis_names.append(axes_dict[keyword]["name"])

        axes_as_expected = np.all(
            np.sort(avail_axis_names) == np.sort(["y", "x"]))
        if axes_as_expected is False:
            print(f"\tIn function {__name__} as expected")
            self.is_valid = False

    def parse(self, hspy_s2d):
        """Parse a hyperspy Signal2D instance into an NX default plottable."""
        if self.is_valid is False:
            pass
        print(f"\tIn function {__name__}")
        self.meta["title"].value = hspy_s2d.metadata["General"]["title"]
        # self.meta["long_name"].value = hspy_s2d.metadata["Signal"]["signal_type"]
        self.meta["long_name"].value = hspy_s2d.metadata["General"]["title"]
        self.meta["counts"].value = hspy_s2d.data    # hspy uses numpy and adapts ??
        axes_dict = hspy_s2d.axes_manager.as_dictionary()
        for keyword in axes_dict.keys():
            offset = np.float64(axes_dict[keyword]["offset"])
            scale = np.float64(axes_dict[keyword]["scale"])
            size = np.uint32(axes_dict[keyword]["size"])
            unit = str(axes_dict[keyword]["units"])
            if axes_dict[keyword]["name"] == "y":
                assert axes_dict[keyword]["_type"] == "UniformDataAxis", \
                    keyword + ", this x axis is not of type UniformDataAxis !"
                self.meta["ypos"].value = np.asarray(
                    np.linspace(0., np.float64(size) * scale, num=size,
                                endpoint=True) + offset / 2., np.float64)
                self.meta["ypos"].unit = unit
                self.meta["ypos_long_name"].value = "y"
            else:  # axes_dict[keyword]["name"] == "x":
                assert axes_dict[keyword]["_type"] == "UniformDataAxis", \
                    keyword + ", this y axis is not of type UniformDataAxis !"
                self.meta["xpos"].value = np.asarray(
                    np.linspace(0., np.float64(size) * scale, num=size,
                                endpoint=True) + offset / 2., np.float64)
                self.meta["xpos"].unit = unit
                self.meta["xpos_long_name"].value = "x"


class NxSpectrumSetEmXray:
    """Representing a set of X-ray spectra with metadata."""

    def __init__(self, hspy_list):
        self.stack_data = []  # the HspyRectRoiXrayAllSpectra
        self.summary_data = []  # the HspyRectRoiXraySummarySpectrum
        self.composition_map = {}  # instances of HspyRectRoiXrayMap
        self.is_valid = True

        self.is_an_implemented_case(hspy_list)
        self.parse_hspy_instances(hspy_list)

    def is_an_implemented_case(self, hspy_list):
        """Check if signal instances in a list is a supported combination."""
        # an Xray analysis which this implementation supports should consist of
        # a rectangular ROI are currently supported, this ROI should have
        # one HAADF overview image
        # this HAADF image is read however by another, a NxImageSetAdf instance
        # of the em parser!
        # one SpectraStack cnts = f(y, x, energy)
        # one accumulated "sum" spectra cnts = f(energy)
        # an arbitrary number of single element X-ray maps cnts = f(y, x)
        cardinality_stack = 0
        cardinality_summary = 0
        for hspy_clss in hspy_list:
            if isinstance(hspy_clss, hs.signals.EDSTEMSpectrum) is True:
                assert hspy_clss.data.ndim in [1, 3], \
                    "Unexpectedly found unsupported-dimensional EDSTEMSpectrum!"
                if hspy_clss.data.ndim == 1:
                    cardinality_summary += 1
                elif hspy_clss.data.ndim == 3:
                    cardinality_stack += 1
                else:
                    continue
        if cardinality_stack != 1:  # stack needed or (cardinality_summary == 1)
            self.is_valid = False

    def parse_hspy_instances(self, hspy_list):
        """Extract from hspy class instances what NOMAD OASIS understands."""
        if self.is_valid is False:
            pass
        print(f"\tIn function {__name__}")
        for hspy_clss in hspy_list:
            if isinstance(hspy_clss, hs.signals.EDSTEMSpectrum) is True:
                ndim = hspy_clss.data.ndim
                if ndim == 1:
                    self.summary_data.append(
                        HspyRectRoiXraySummarySpectrum(hspy_clss))
                elif ndim == 3:
                    self.stack_data.append(
                        HspyRectRoiXrayAllSpectra(hspy_clss))
                else:
                    continue
            elif isinstance(hspy_clss, hs.signals.Signal2D) is True:
                ndim = hspy_clss.data.ndim
                if ndim == 2:
                    title = hspy_clss.metadata["General"]["title"]
                    if title != "HAADF":
                        self.composition_map[title] \
                            = HspyRectRoiXrayMap(hspy_clss)
            else:
                continue

    def report(self, prefix: str, frame_id: int,
               ifo: dict, template: dict) -> dict:
        """Enter data from the NX-specific representation into the template."""
        if self.is_valid is False:
            print(f"\t{__name__} reporting nothing!")
            return template
        print(f"\t{__name__} reporting...")
        assert (len(self.stack_data) >= 0) and (len(self.stack_data) <= 1), \
            "More than one spectrum stack is currently not supported!"
        assert (len(self.summary_data) >= 0) and (len(self.summary_data) <= 1), \
            "More than one sum spectrum stack is currently not supported!"
        # for keyword, obj in self.composition_map.items():
        #     print(f"{keyword}, np.shape(obj.counts.value), {np.shape(obj.counts.value)}")

        if len(self.stack_data) == 1:
            trg = f"{prefix}xray/PROCESS[process1]/"
            template[f"{trg}PROGRAM[program1]/program"] = "hyperspy"
            template[f"{trg}PROGRAM[program1]/program/@version"] \
                = hs.__version__
            template[f"{trg}mode"] = "n/a"
            template[f"{trg}detector_identifier"] = "n/a"
            template[f"{trg}source"] = ifo["source_file_name"]
            template[f"{trg}source/@version"] = ifo["source_file_version"]

            # MISSING_DATA_MSG = "not in hspy metadata case specifically in original_metadata"
            # ##MK::!!!!!
            # ##MK::how to communicate that these data do not exist?
            # template[trg + "adf_outer_half_angle"] = np.float64(0.)
            # template[trg + "adf_outer_half_angle/@units"] = "rad"
            # template[trg + "adf_inner_half_angle"] = np.float64(0.)
            # template[trg + "adf_inner_half_angle/@units"] = "rad"

            trg = f"{prefix}xray/stack/"
            template[f"{trg}title"] = "Xray spectra stack"
            # template[f"{trg}@long_name"] \
            #     = self.stack_data[0].meta["long_name"].value
            template[f"{trg}@signal"] = "data_counts"
            template[f"{trg}@axes"] \
                = ["axis_y", "axis_x", "axis_photon_energy"]
            template[f"{trg}@AXISNAME_indices[axis_photon_energy_indices]"] \
                = np.uint32(2)
            template[f"{trg}@AXISNAME_indices[axis_x_indices]"] = np.uint32(1)
            template[f"{trg}@AXISNAME_indices[axis_y_indices]"] = np.uint32(0)
            template[f"{trg}DATA[data_counts]"] \
                = {"compress": self.stack_data[0].meta["counts"].value,
                   "strength": 1}
            template[f"{trg}DATA[data_counts]/@long_name"] = "Photon counts (1)"
            template[f"{trg}AXISNAME[axis_photon_energy]"] \
                = {"compress": self.stack_data[0].meta["photon_energy"].value,
                   "strength": 1}
            template[f"{trg}AXISNAME[axis_photon_energy]/@units"] \
                = self.stack_data[0].meta["photon_energy"].unit
            template[f"{trg}AXISNAME[axis_photon_energy]/@long_name"] \
                = f"Photon energy ({self.stack_data[0].meta['photon_energy'].unit})"
            template[f"{trg}AXISNAME[axis_x]"] \
                = {"compress": self.stack_data[0].meta["xpos"].value,
                   "strength": 1}
            template[f"{trg}AXISNAME[axis_x]/@units"] \
                = self.stack_data[0].meta["xpos"].unit
            template[f"{trg}AXISNAME[axis_x]/@long_name"] \
                = f"x ({self.stack_data[0].meta['xpos'].unit})"
            template[f"{trg}AXISNAME[axis_y]"] \
                = {"compress": self.stack_data[0].meta["ypos"].value,
                   "strength": 1}
            template[f"{trg}AXISNAME[axis_y]/@units"] \
                = self.stack_data[0].meta["ypos"].unit
            template[f"{trg}AXISNAME[axis_y]/@long_name"] \
                = f"y ({self.stack_data[0].meta['ypos'].unit})"

        if len(self.summary_data) == 1:
            trg = f"{prefix}xray/summary/"
            template[f"{trg}title"] = "Accumulated X-ray spectrum"
            # template[f"{trg}@long_name"] = self.summary_data[0].meta["long_name"].value
            template[f"{trg}@signal"] = "data_counts"
            template[f"{trg}@axes"] = ["axis_photon_energy"]
            template[f"{trg}@AXISNAME_indices[axis_photon_energy_indices]"] \
                = np.uint32(0)
            template[f"{trg}DATA[data_counts]"] \
                = {"compress": self.summary_data[0].meta["counts"].value,
                   "strength": 1}
            template[f"{trg}DATA[data_counts]/@long_name"] = "Photon counts (1)"
            template[f"{trg}AXISNAME[axis_photon_energy]"] \
                = {"compress": self.summary_data[0].meta["photon_energy"].value,
                   "strength": 1}
            template[f"{trg}AXISNAME[axis_photon_energy]/@units"] \
                = self.summary_data[0].meta["photon_energy"].unit
            template[f"{trg}AXISNAME[axis_photon_energy]/@long_name"] \
                = f"Photon energy ({self.summary_data[0].meta['photon_energy'].unit})"

        return template

        # skip the composition maps for the search sprint
        # for keyword, xray_map in self.composition_map.items():
        #     prfx = f"{trg}PROCESS[indexing]/PROCESS[{keyword.lower()}]/summary/"
        #     template[f"{prfx}title"] = f"X-ray mapping for {keyword}"
        #     # = xray_map.meta["long_name"].value
        #     # template[f"{prfx}@long_name"] = xray_map.meta["long_name"].value
        #     template[f"{prfx}@signal"] = "data_counts"
        #     template[f"{prfx}@axes"] = ["axis_y", "axis_x"]
        #     template[f"{prfx}@AXISNAME[axis_x_indices]"] = np.uint32(1)
        #     template[f"{prfx}@AXISNAME[axis_y_indices]"] = np.uint32(0)
        #     template[f"{prfx}DATA[data_counts]"] \
        #         = {"compress": xray_map.meta["counts"].value, "strength": 1}
        #     template[f"{prfx}DATA[data_counts]/@long_name"] = "Counts (1)"
        #     template[f"{prfx}AXISNAME[axis_x]"] \
        #         = {"compress": xray_map.meta["xpos"].value, "strength": 1}
        #     template[f"{prfx}AXISNAME[axis_x]/@units"] = xray_map.meta["xpos"].unit
        #     template[f"{prfx}AXISNAME[axis_x]/@long_name"] \
        #         = f"x ({xray_map.meta['xpos'].unit})"
        #     template[f"{prfx}AXISNAME[axis_y]"] \
        #         = {"compress": xray_map.meta["ypos"].value, "strength": 1}
        #     template[f"{prfx}AXISNAME[axis_y]/@units"] = xray_map.meta["ypos"].unit
        #     template[f"{prfx}AXISNAME[axis_y]/@long_name"] \
        #         = f"y ({xray_map.meta['ypos'].unit})"



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Parser for loading generic X-ray spectroscopy and EELS data into NXem via hyperspy."""

# pylint: disable=no-member

from typing import Tuple, Any

from pynxtools.dataconverter.readers.base.reader import BaseReader

from pynxtools.dataconverter.readers.em_spctrscpy.utils.em_use_case_selector \
    import EmUseCaseSelector

from pynxtools.dataconverter.readers.em_spctrscpy.utils.em_generic_eln_io \
    import NxEmNomadOasisElnSchemaParser

from pynxtools.dataconverter.readers.em_spctrscpy.utils.em_event_data \
    import NxEventDataEm

from pynxtools.dataconverter.readers.em_spctrscpy.utils.em_nexus_plots \
    import em_spctrscpy_default_plot_generator

from pynxtools.dataconverter.readers.em_spctrscpy.utils.em_example_data \
    import EmSpctrscpyCreateExampleData


def hyperspy_parser(file_name: str, template: dict, entry_id: int) -> dict:
    """Parse content from electron microscopy vendor files."""
    test = NxEventDataEm(file_name, entry_id)
    test.report(template)
    return template


def nomad_oasis_eln_parser(file_name: str, template: dict, entry_id: int) -> dict:
    """Parse out output from a YAML file from a NOMAD OASIS YAML."""
    test = NxEmNomadOasisElnSchemaParser(file_name, entry_id)
    test.report(template)
    return template


class EmSpctrscpyReader(BaseReader):
    """Parse content from community file formats.

    Specifically, electron microscopy
    towards a NXem.nxdl-compliant NeXus file.
    """

    # pylint: disable=too-few-public-methods

    # Whitelist for the NXDLs that the reader supports and can process
    supported_nxdls = ["NXem"]

    # pylint: disable=duplicate-code
    def read(self,
             template: dict = None,
             file_paths: Tuple[str] = None,
             objects: Tuple[Any] = None) -> dict:
        """Read data from given file, return filled template dictionary em."""
        # pylint: disable=duplicate-code
        template.clear()

        # this em_spctrscpy parser combines multiple sub-parsers
        # so we need the following input:
        # logical analysis which use case
        # data input from an ELN (e.g. NOMAD OASIS)
        # data input from technology partner files
        # functionalities for creating default plots
        # developer functionalities for creating synthetic data

        n_entries = 1
        entry_id = 1
        if len(file_paths) == 1:
            if file_paths[0].startswith("synthesize"):
                synthesis_id = int(file_paths[0].replace("synthesize", ""))
                print(f"synthesis_id {synthesis_id}")
            else:
                synthesis_id = 1
            print("Create one synthetic entry in one NeXus file...")
            synthetic = EmSpctrscpyCreateExampleData(synthesis_id)
            synthetic.synthesize(template)
        else:
            print("Parse ELN and technology partner file(s)...")
            case = EmUseCaseSelector(file_paths)
            assert case.is_valid is True, \
                "Such a combination of input-file(s, if any) is not supported !"

            print("Parse (meta)data coming from an ELN...")
            if case.eln_parser == "nomad-oasis":
                nomad_oasis_eln_parser(case.eln[0], template, entry_id)
            else:
                print("No input file defined for eln data !")
                return {}

            print("Parse (numerical) data and metadata from technology partner files...")
            if case.vendor_parser == "oina":
                # oina_parser(case.vendor[0], template, entry_id)
                return {}
            if case.vendor_parser == "hspy":
                hyperspy_parser(case.vendor[0], template, entry_id)
            else:
                print("No input-file defined for technology partner data !")
                return {}

        print("Create NeXus default plottable data...")
        em_spctrscpy_default_plot_generator(template, n_entries)

        debugging = False  # True
        if debugging is True:
            print("Reporting state of template before passing to HDF5 writing...")
            for keyword in template.keys():
                print(keyword)
                # print(type(template[keyword]))
                # print(template[keyword])

        print("Forward instantiated template to the NXS writer...")
        return template


# This has to be set to allow the convert script to use this reader.
READER = EmSpctrscpyReader



"""
Build wrapper for setuptools to create a nexus-version.txt file
containing the nexus definitions verison.
"""
import os
from subprocess import CalledProcessError, run
from typing import Optional

from setuptools import build_meta as _orig
from setuptools.build_meta import *  # pylint: disable=wildcard-import,unused-wildcard-import


def get_vcs_version(tag_match="*[0-9]*") -> Optional[str]:
    """
    The version of the Nexus standard and the NeXus Definition language
    based on git tags and commits
    """
    try:
        return (
            run(
                [
                    "git",
                    "describe",
                    "--dirty",
                    "--tags",
                    "--long",
                    "--match",
                    tag_match,
                ],
                cwd=os.path.join(os.path.dirname(__file__), "../pynxtools/definitions"),
                check=True,
                capture_output=True,
            )
            .stdout.decode("utf-8")
            .strip()
        )
    except (FileNotFoundError, CalledProcessError):
        return None


def _write_version_to_metadata():
    version = get_vcs_version()
    if version is None or not version:
        return

    with open(
        os.path.join(os.path.dirname(__file__), "nexus-version.txt"),
        "w+",
        encoding="utf-8",
    ) as file:
        file.write(version)


# pylint: disable=function-redefined
def build_wheel(wheel_directory, config_settings=None, metadata_directory=None):
    """
    PEP 517 compliant build wheel hook.
    This is a wrapper for setuptools and adds a nexus version file.
    """
    _write_version_to_metadata()
    return _orig.build_wheel(wheel_directory, config_settings, metadata_directory)


# pylint: disable=function-redefined
def build_sdist(sdist_directory, config_settings=None):
    """
    PEP 517 compliant build sdist hook.
    This is a wrapper for setuptools and adds a nexus version file.
    """
    _write_version_to_metadata()
    return _orig.build_sdist(sdist_directory, config_settings)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Definitions for right- or left-handed Cartesian coordinate system."""

# pylint: disable=no-member

# the nxebsd.schema.archive.yaml" offers a set of sets of RadioEditButtons
# each such set can be used to define the unique orientation of a right-handed
# Cartesian coordinate system's base vectors when viewed by an outer observer
# looking to the display of the coordinate system triplet vector and comparing
# to the plane of sight and a compass
# the compass defines four directions, north aka up, east aka right, south aka down,
# and west aka left. The plane of sight defines two additional directions in aka
# into the plane and out aka out-of-the-plane
# there are different strategies what users find convenient to report to specify
# the directions into which the base vectors are assumed pointing
# in general only two perpendicular directions have to be specified as the third
# follows implicitly i.e. there is no need to overconstraint hence the term "undefined"
# the following dictionary is a lookup table whereby the input of the user
# can be evaluated to check if the radioeditbutton choice result in a fully constraint
# and thus unique definition. All other choices are reported as a not fully constraint
# i.e. eventually problematic coordinate system

REFERENCE_FRAMES = ["undefined", "right_handed_cartesian", "left_handed_cartesian"]

AXIS_DIRECTIONS = ["undefined", "north", "east", "south", "west", "in", "out"]

# is a right-handed Cartesian coordinate system sufficiently constrained, when
# at least two base vector directions are chosen
is_cs_rh_unambiguous = {"undefined_undefined_undefined": False,
                        "undefined_undefined_north": False,
                        "undefined_undefined_east": False,
                        "undefined_undefined_south": False,
                        "undefined_undefined_west": False,
                        "undefined_undefined_in": False,
                        "undefined_undefined_out": False,
                        "undefined_north_undefined": False,
                        "undefined_north_north": False,
                        "undefined_north_east": True,
                        "undefined_north_south": True,
                        "undefined_north_west": True,
                        "undefined_north_in": True,
                        "undefined_north_out": True,
                        "undefined_east_undefined": False,
                        "undefined_east_north": True,
                        "undefined_east_east": False,
                        "undefined_east_south": True,
                        "undefined_east_west": False,
                        "undefined_east_in": True,
                        "undefined_east_out": True,
                        "undefined_south_undefined": False,
                        "undefined_south_north": False,
                        "undefined_south_east": True,
                        "undefined_south_south": False,
                        "undefined_south_west": True,
                        "undefined_south_in": True,
                        "undefined_south_out": True,
                        "undefined_west_undefined": False,
                        "undefined_west_north": True,
                        "undefined_west_east": False,
                        "undefined_west_south": True,
                        "undefined_west_west": False,
                        "undefined_west_in": True,
                        "undefined_west_out": True,
                        "undefined_in_undefined": False,
                        "undefined_in_north": True,
                        "undefined_in_east": True,
                        "undefined_in_south": True,
                        "undefined_in_west": True,
                        "undefined_in_in": False,
                        "undefined_in_out": False,
                        "undefined_out_undefined": False,
                        "undefined_out_north": True,
                        "undefined_out_east": True,
                        "undefined_out_south": True,
                        "undefined_out_west": True,
                        "undefined_out_in": False,
                        "undefined_out_out": False,
                        "north_undefined_undefined": False,
                        "north_undefined_north": False,
                        "north_undefined_east": True,
                        "north_undefined_south": False,
                        "north_undefined_west": True,
                        "north_undefined_in": True,
                        "north_undefined_out": True,
                        "north_north_undefined": False,
                        "north_north_north": False,
                        "north_north_east": False,
                        "north_north_south": False,
                        "north_north_west": False,
                        "north_north_in": False,
                        "north_north_out": False,
                        "north_east_undefined": True,
                        "north_east_north": False,
                        "north_east_east": False,
                        "north_east_south": False,
                        "north_east_west": False,
                        "north_east_in": True,
                        "north_east_out": False,
                        "north_south_undefined": False,
                        "north_south_north": False,
                        "north_south_east": False,
                        "north_south_south": False,
                        "north_south_west": False,
                        "north_south_in": False,
                        "north_south_out": False,
                        "north_west_undefined": True,
                        "north_west_north": False,
                        "north_west_east": False,
                        "north_west_south": False,
                        "north_west_west": False,
                        "north_west_in": False,
                        "north_west_out": True,
                        "north_in_undefined": True,
                        "north_in_north": False,
                        "north_in_east": False,
                        "north_in_south": False,
                        "north_in_west": True,
                        "north_in_in": False,
                        "north_in_out": False,
                        "north_out_undefined": True,
                        "north_out_north": False,
                        "north_out_east": True,
                        "north_out_south": False,
                        "north_out_west": False,
                        "north_out_in": False,
                        "north_out_out": False,
                        "east_undefined_undefined": False,
                        "east_undefined_north": True,
                        "east_undefined_east": False,
                        "east_undefined_south": True,
                        "east_undefined_west": False,
                        "east_undefined_in": True,
                        "east_undefined_out": True,
                        "east_north_undefined": True,
                        "east_north_north": False,
                        "east_north_east": False,
                        "east_north_south": False,
                        "east_north_west": False,
                        "east_north_in": False,
                        "east_north_out": True,
                        "east_east_undefined": False,
                        "east_east_north": False,
                        "east_east_east": False,
                        "east_east_south": False,
                        "east_east_west": False,
                        "east_east_in": False,
                        "east_east_out": False,
                        "east_south_undefined": True,
                        "east_south_north": False,
                        "east_south_east": False,
                        "east_south_south": False,
                        "east_south_west": False,
                        "east_south_in": True,
                        "east_south_out": False,
                        "east_west_undefined": False,
                        "east_west_north": False,
                        "east_west_east": False,
                        "east_west_south": False,
                        "east_west_west": False,
                        "east_west_in": False,
                        "east_west_out": False,
                        "east_in_undefined": True,
                        "east_in_north": True,
                        "east_in_east": False,
                        "east_in_south": False,
                        "east_in_west": False,
                        "east_in_in": False,
                        "east_in_out": False,
                        "east_out_undefined": True,
                        "east_out_north": False,
                        "east_out_east": False,
                        "east_out_south": True,
                        "east_out_west": False,
                        "east_out_in": False,
                        "east_out_out": False,
                        "south_undefined_undefined": False,
                        "south_undefined_north": False,
                        "south_undefined_east": True,
                        "south_undefined_south": False,
                        "south_undefined_west": True,
                        "south_undefined_in": True,
                        "south_undefined_out": True,
                        "south_north_undefined": False,
                        "south_north_north": False,
                        "south_north_east": False,
                        "south_north_south": False,
                        "south_north_west": False,
                        "south_north_in": False,
                        "south_north_out": False,
                        "south_east_undefined": True,
                        "south_east_north": False,
                        "south_east_east": False,
                        "south_east_south": False,
                        "south_east_west": False,
                        "south_east_in": False,
                        "south_east_out": True,
                        "south_south_undefined": False,
                        "south_south_north": False,
                        "south_south_east": False,
                        "south_south_south": False,
                        "south_south_west": False,
                        "south_south_in": False,
                        "south_south_out": False,
                        "south_west_undefined": True,
                        "south_west_north": False,
                        "south_west_east": False,
                        "south_west_south": False,
                        "south_west_west": False,
                        "south_west_in": True,
                        "south_west_out": False,
                        "south_in_undefined": True,
                        "south_in_north": False,
                        "south_in_east": True,
                        "south_in_south": False,
                        "south_in_west": False,
                        "south_in_in": False,
                        "south_in_out": False,
                        "south_out_undefined": True,
                        "south_out_north": False,
                        "south_out_east": False,
                        "south_out_south": False,
                        "south_out_west": True,
                        "south_out_in": False,
                        "south_out_out": False,
                        "west_undefined_undefined": False,
                        "west_undefined_north": True,
                        "west_undefined_east": False,
                        "west_undefined_south": True,
                        "west_undefined_west": False,
                        "west_undefined_in": True,
                        "west_undefined_out": True,
                        "west_north_undefined": True,
                        "west_north_north": False,
                        "west_north_east": False,
                        "west_north_south": False,
                        "west_north_west": False,
                        "west_north_in": True,
                        "west_north_out": False,
                        "west_east_undefined": False,
                        "west_east_north": False,
                        "west_east_east": False,
                        "west_east_south": False,
                        "west_east_west": False,
                        "west_east_in": False,
                        "west_east_out": False,
                        "west_south_undefined": True,
                        "west_south_north": False,
                        "west_south_east": False,
                        "west_south_south": False,
                        "west_south_west": False,
                        "west_south_in": False,
                        "west_south_out": True,
                        "west_west_undefined": False,
                        "west_west_north": False,
                        "west_west_east": False,
                        "west_west_south": False,
                        "west_west_west": False,
                        "west_west_in": False,
                        "west_west_out": False,
                        "west_in_undefined": True,
                        "west_in_north": False,
                        "west_in_east": False,
                        "west_in_south": True,
                        "west_in_west": False,
                        "west_in_in": False,
                        "west_in_out": False,
                        "west_out_undefined": True,
                        "west_out_north": True,
                        "west_out_east": False,
                        "west_out_south": False,
                        "west_out_west": False,
                        "west_out_in": False,
                        "west_out_out": False,
                        "in_undefined_undefined": False,
                        "in_undefined_north": True,
                        "in_undefined_east": True,
                        "in_undefined_south": True,
                        "in_undefined_west": True,
                        "in_undefined_in": False,
                        "in_undefined_out": False,
                        "in_north_undefined": True,
                        "in_north_north": False,
                        "in_north_east": True,
                        "in_north_south": False,
                        "in_north_west": False,
                        "in_north_in": False,
                        "in_north_out": False,
                        "in_east_undefined": True,
                        "in_east_north": False,
                        "in_east_east": False,
                        "in_east_south": True,
                        "in_east_west": False,
                        "in_east_in": False,
                        "in_east_out": False,
                        "in_south_undefined": True,
                        "in_south_north": False,
                        "in_south_east": False,
                        "in_south_south": False,
                        "in_south_west": True,
                        "in_south_in": False,
                        "in_south_out": False,
                        "in_west_undefined": True,
                        "in_west_north": True,
                        "in_west_east": False,
                        "in_west_south": False,
                        "in_west_west": False,
                        "in_west_in": False,
                        "in_west_out": False,
                        "in_in_undefined": False,
                        "in_in_north": False,
                        "in_in_east": False,
                        "in_in_south": False,
                        "in_in_west": False,
                        "in_in_in": False,
                        "in_in_out": False,
                        "in_out_undefined": False,
                        "in_out_north": False,
                        "in_out_east": False,
                        "in_out_south": False,
                        "in_out_west": False,
                        "in_out_in": False,
                        "in_out_out": False,
                        "out_undefined_undefined": False,
                        "out_undefined_north": True,
                        "out_undefined_east": True,
                        "out_undefined_south": True,
                        "out_undefined_west": True,
                        "out_undefined_in": False,
                        "out_undefined_out": False,
                        "out_north_undefined": True,
                        "out_north_north": False,
                        "out_north_east": False,
                        "out_north_south": False,
                        "out_north_west": True,
                        "out_north_in": False,
                        "out_north_out": False,
                        "out_east_undefined": True,
                        "out_east_north": True,
                        "out_east_east": False,
                        "out_east_south": False,
                        "out_east_west": False,
                        "out_east_in": False,
                        "out_east_out": False,
                        "out_south_undefined": True,
                        "out_south_north": False,
                        "out_south_east": True,
                        "out_south_south": False,
                        "out_south_west": False,
                        "out_south_in": False,
                        "out_south_out": False,
                        "out_west_undefined": True,
                        "out_west_north": False,
                        "out_west_east": False,
                        "out_west_south": True,
                        "out_west_west": False,
                        "out_west_in": False,
                        "out_west_out": False,
                        "out_in_undefined": False,
                        "out_in_north": False,
                        "out_in_east": False,
                        "out_in_south": False,
                        "out_in_west": False,
                        "out_in_in": False,
                        "out_in_out": False,
                        "out_out_undefined": False,
                        "out_out_north": False,
                        "out_out_east": False,
                        "out_out_south": False,
                        "out_out_west": False,
                        "out_out_in": False,
                        "out_out_out": False}

is_cs_lh_unambiguous = {"undefined_undefined_undefined": False,
                        "undefined_undefined_north": False,
                        "undefined_undefined_east": False,
                        "undefined_undefined_south": False,
                        "undefined_undefined_west": False,
                        "undefined_undefined_in": False,
                        "undefined_undefined_out": False,
                        "undefined_north_undefined": False,
                        "undefined_north_north": False,
                        "undefined_north_east": True,
                        "undefined_north_south": False,
                        "undefined_north_west": True,
                        "undefined_north_in": True,
                        "undefined_north_out": True,
                        "undefined_east_undefined": False,
                        "undefined_east_north": True,
                        "undefined_east_east": False,
                        "undefined_east_south": True,
                        "undefined_east_west": False,
                        "undefined_east_in": True,
                        "undefined_east_out": True,
                        "undefined_south_undefined": False,
                        "undefined_south_north": False,
                        "undefined_south_east": True,
                        "undefined_south_south": False,
                        "undefined_south_west": True,
                        "undefined_south_in": True,
                        "undefined_south_out": True,
                        "undefined_west_undefined": False,
                        "undefined_west_north": True,
                        "undefined_west_east": False,
                        "undefined_west_south": True,
                        "undefined_west_west": False,
                        "undefined_west_in": True,
                        "undefined_west_out": True,
                        "undefined_in_undefined": False,
                        "undefined_in_north": True,
                        "undefined_in_east": True,
                        "undefined_in_south": True,
                        "undefined_in_west": True,
                        "undefined_in_in": False,
                        "undefined_in_out": False,
                        "undefined_out_undefined": False,
                        "undefined_out_north": True,
                        "undefined_out_east": True,
                        "undefined_out_south": True,
                        "undefined_out_west": True,
                        "undefined_out_in": False,
                        "undefined_out_out": False,
                        "north_undefined_undefined": False,
                        "north_undefined_north": False,
                        "north_undefined_east": True,
                        "north_undefined_south": False,
                        "north_undefined_west": True,
                        "north_undefined_in": True,
                        "north_undefined_out": True,
                        "north_north_undefined": False,
                        "north_north_north": False,
                        "north_north_east": False,
                        "north_north_south": False,
                        "north_north_west": False,
                        "north_north_in": False,
                        "north_north_out": False,
                        "north_east_undefined": True,
                        "north_east_north": False,
                        "north_east_east": False,
                        "north_east_south": False,
                        "north_east_west": False,
                        "north_east_in": False,
                        "north_east_out": True,
                        "north_south_undefined": False,
                        "north_south_north": False,
                        "north_south_east": False,
                        "north_south_south": False,
                        "north_south_west": False,
                        "north_south_in": False,
                        "north_south_out": False,
                        "north_west_undefined": True,
                        "north_west_north": False,
                        "north_west_east": False,
                        "north_west_south": False,
                        "north_west_west": False,
                        "north_west_in": True,
                        "north_west_out": False,
                        "north_in_undefined": True,
                        "north_in_north": False,
                        "north_in_east": True,
                        "north_in_south": False,
                        "north_in_west": False,
                        "north_in_in": False,
                        "north_in_out": False,
                        "north_out_undefined": True,
                        "north_out_north": False,
                        "north_out_east": False,
                        "north_out_south": False,
                        "north_out_west": True,
                        "north_out_in": False,
                        "north_out_out": False,
                        "east_undefined_undefined": False,
                        "east_undefined_north": True,
                        "east_undefined_east": False,
                        "east_undefined_south": True,
                        "east_undefined_west": False,
                        "east_undefined_in": True,
                        "east_undefined_out": True,
                        "east_north_undefined": True,
                        "east_north_north": False,
                        "east_north_east": False,
                        "east_north_south": False,
                        "east_north_west": False,
                        "east_north_in": True,
                        "east_north_out": False,
                        "east_east_undefined": False,
                        "east_east_north": False,
                        "east_east_east": False,
                        "east_east_south": False,
                        "east_east_west": False,
                        "east_east_in": False,
                        "east_east_out": False,
                        "east_south_undefined": True,
                        "east_south_north": False,
                        "east_south_east": False,
                        "east_south_south": False,
                        "east_south_west": False,
                        "east_south_in": False,
                        "east_south_out": True,
                        "east_west_undefined": False,
                        "east_west_north": False,
                        "east_west_east": False,
                        "east_west_south": False,
                        "east_west_west": False,
                        "east_west_in": False,
                        "east_west_out": False,
                        "east_in_undefined": True,
                        "east_in_north": False,
                        "east_in_east": False,
                        "east_in_south": True,
                        "east_in_west": False,
                        "east_in_in": False,
                        "east_in_out": False,
                        "east_out_undefined": True,
                        "east_out_north": True,
                        "east_out_east": False,
                        "east_out_south": False,
                        "east_out_west": False,
                        "east_out_in": False,
                        "east_out_out": False,
                        "south_undefined_undefined": False,
                        "south_undefined_north": False,
                        "south_undefined_east": True,
                        "south_undefined_south": False,
                        "south_undefined_west": True,
                        "south_undefined_in": True,
                        "south_undefined_out": True,
                        "south_north_undefined": False,
                        "south_north_north": False,
                        "south_north_east": False,
                        "south_north_south": False,
                        "south_north_west": False,
                        "south_north_in": False,
                        "south_north_out": False,
                        "south_east_undefined": True,
                        "south_east_north": False,
                        "south_east_east": False,
                        "south_east_south": False,
                        "south_east_west": False,
                        "south_east_in": True,
                        "south_east_out": False,
                        "south_south_undefined": False,
                        "south_south_north": False,
                        "south_south_east": False,
                        "south_south_south": False,
                        "south_south_west": False,
                        "south_south_in": False,
                        "south_south_out": False,
                        "south_west_undefined": True,
                        "south_west_north": False,
                        "south_west_east": False,
                        "south_west_south": False,
                        "south_west_west": False,
                        "south_west_in": False,
                        "south_west_out": True,
                        "south_in_undefined": True,
                        "south_in_north": False,
                        "south_in_east": False,
                        "south_in_south": False,
                        "south_in_west": True,
                        "south_in_in": False,
                        "south_in_out": False,
                        "south_out_undefined": True,
                        "south_out_north": False,
                        "south_out_east": True,
                        "south_out_south": False,
                        "south_out_west": False,
                        "south_out_in": False,
                        "south_out_out": False,
                        "west_undefined_undefined": False,
                        "west_undefined_north": True,
                        "west_undefined_east": False,
                        "west_undefined_south": True,
                        "west_undefined_west": False,
                        "west_undefined_in": True,
                        "west_undefined_out": True,
                        "west_north_undefined": True,
                        "west_north_north": False,
                        "west_north_east": False,
                        "west_north_south": False,
                        "west_north_west": False,
                        "west_north_in": False,
                        "west_north_out": True,
                        "west_east_undefined": False,
                        "west_east_north": False,
                        "west_east_east": False,
                        "west_east_south": False,
                        "west_east_west": False,
                        "west_east_in": False,
                        "west_east_out": False,
                        "west_south_undefined": True,
                        "west_south_north": False,
                        "west_south_east": False,
                        "west_south_south": False,
                        "west_south_west": False,
                        "west_south_in": True,
                        "west_south_out": False,
                        "west_west_undefined": False,
                        "west_west_north": False,
                        "west_west_east": False,
                        "west_west_south": False,
                        "west_west_west": False,
                        "west_west_in": False,
                        "west_west_out": False,
                        "west_in_undefined": True,
                        "west_in_north": True,
                        "west_in_east": False,
                        "west_in_south": False,
                        "west_in_west": False,
                        "west_in_in": False,
                        "west_in_out": False,
                        "west_out_undefined": True,
                        "west_out_north": False,
                        "west_out_east": False,
                        "west_out_south": True,
                        "west_out_west": False,
                        "west_out_in": False,
                        "west_out_out": False,
                        "in_undefined_undefined": False,
                        "in_undefined_north": True,
                        "in_undefined_east": True,
                        "in_undefined_south": True,
                        "in_undefined_west": True,
                        "in_undefined_in": False,
                        "in_undefined_out": False,
                        "in_north_undefined": True,
                        "in_north_north": False,
                        "in_north_east": False,
                        "in_north_south": False,
                        "in_north_west": True,
                        "in_north_in": False,
                        "in_north_out": False,
                        "in_east_undefined": True,
                        "in_east_north": True,
                        "in_east_east": False,
                        "in_east_south": False,
                        "in_east_west": False,
                        "in_east_in": False,
                        "in_east_out": False,
                        "in_south_undefined": True,
                        "in_south_north": False,
                        "in_south_east": True,
                        "in_south_south": False,
                        "in_south_west": False,
                        "in_south_in": False,
                        "in_south_out": False,
                        "in_west_undefined": True,
                        "in_west_north": False,
                        "in_west_east": False,
                        "in_west_south": True,
                        "in_west_west": False,
                        "in_west_in": False,
                        "in_west_out": False,
                        "in_in_undefined": False,
                        "in_in_north": False,
                        "in_in_east": False,
                        "in_in_south": False,
                        "in_in_west": False,
                        "in_in_in": False,
                        "in_in_out": False,
                        "in_out_undefined": False,
                        "in_out_north": False,
                        "in_out_east": False,
                        "in_out_south": False,
                        "in_out_west": False,
                        "in_out_in": False,
                        "in_out_out": False,
                        "out_undefined_undefined": False,
                        "out_undefined_north": True,
                        "out_undefined_east": True,
                        "out_undefined_south": True,
                        "out_undefined_west": True,
                        "out_undefined_in": False,
                        "out_undefined_out": False,
                        "out_north_undefined": True,
                        "out_north_north": False,
                        "out_north_east": True,
                        "out_north_south": False,
                        "out_north_west": False,
                        "out_north_in": False,
                        "out_north_out": False,
                        "out_east_undefined": True,
                        "out_east_north": False,
                        "out_east_east": False,
                        "out_east_south": True,
                        "out_east_west": False,
                        "out_east_in": False,
                        "out_east_out": False,
                        "out_south_undefined": True,
                        "out_south_north": False,
                        "out_south_east": False,
                        "out_south_south": False,
                        "out_south_west": True,
                        "out_south_in": False,
                        "out_south_out": False,
                        "out_west_undefined": True,
                        "out_west_north": True,
                        "out_west_east": False,
                        "out_west_south": False,
                        "out_west_west": False,
                        "out_west_in": False,
                        "out_west_out": False,
                        "out_in_undefined": False,
                        "out_in_north": False,
                        "out_in_east": False,
                        "out_in_south": False,
                        "out_in_west": False,
                        "out_in_in": False,
                        "out_in_out": False,
                        "out_out_undefined": False,
                        "out_out_north": False,
                        "out_out_east": False,
                        "out_out_south": False,
                        "out_out_west": False,
                        "out_out_in": False,
                        "out_out_out": False}


def is_cs_well_defined(handedness, directions):
    """Check if the axis directions yield an unambiguous definition right handed"""
    keyword = f"{directions[0]}_{directions[1]}_{directions[2]}"
    if handedness == "right_handed_cartesian":
        if keyword in is_cs_rh_unambiguous:
            if is_cs_rh_unambiguous[keyword] is True:
                return True
    if handedness == "left_handed_cartesian":
        if keyword in is_cs_lh_unambiguous:
            if is_cs_lh_unambiguous[keyword] is True:
                return True
    return False



#!/usr/bin/python

import sys
import nxs
import numpy

a = numpy.zeros((2,3,4),dtype=numpy.int)
val = 0
for i in range(2):
    for j in range(3):
        for k in range(4):
            a[i,j,k] = val
            val = val + 1

nf = nxs.open("simple3D.h5", "w5")

nf.makegroup("entry","NXentry")
nf.opengroup("entry","NXentry")

nf.makegroup("data","NXdata")
nf.opengroup("data","NXdata")
nf.putattr("signal","test")

nf.makedata("test",'int32',[2,3,4])
nf.opendata("test")
nf.putdata(a)
nf.closedata()

nf.closegroup()	# NXdata
nf.closegroup() # NXentry

nf.close()

exit
